from .sdk import EmpyrealSDK
from .types import Application, Network, Token, TokenAmount, User, Wallet


__all__ = [
    "EmpyrealSDK",
    "Application",
    "Network",
    "Token",
    "TokenAmount",
    "User",
    "Wallet",
]



from datetime import datetime
from enum import Enum, auto
from functools import singledispatchmethod
from typing import Optional

from eth_typing import ChecksumAddress, HexAddress, HexStr
from pydantic import BaseModel

from .token import Token, TokenAmount
from .wallet import Wallet
from .network import Network
from ..utils.client import _force_get_global_client


class Liquidity(BaseModel):
    token0_balance: int
    token0_price: Optional[float] = None
    token1_balance: int
    token1_price: Optional[float] = None
    pair: "DexPair"

    @property
    def token0_value(self):
        # TODO: handle invalid prices
        if self.token0_price:
            return (
                self.token0_price
                * self.token0_balance
                / 10**self.pair.token0.decimals
            )
        return -0.5

    @property
    def token1_value(self):
        # TODO: handle invalid prices
        if self.token1_price:
            return (
                self.token1_price
                * self.token1_balance
                / 10**self.pair.token1.decimals
            )
        return -0.5

    @property
    def value(self):
        return round(self.token0_value * 2, 2)

    def __repr__(self):
        token0_symbol = self.pair.token0.symbol
        token1_symbol = self.pair.token1.symbol
        token0 = self.token0_balance / 10**self.pair.token0.decimals
        token1 = self.token1_balance / 10**self.pair.token1.decimals
        return f"<Liquidity: ${format(self.value, ',')} ({token0_symbol}: {token0}, {token1_symbol}: {token1})>"

    __str__ = __repr__


class DexRoute(BaseModel):
    path: list[ChecksumAddress]
    pair_addresses: list[ChecksumAddress]
    eth_price: TokenAmount
    usdc_price: TokenAmount
    factory: "DexFactory"

    def __repr__(self):
        return f"<DexRoute | path={self.path} | eth_price: {self.eth_price.format(8)}, usdc_price: ${self.usdc_price.format(6)}>"

    __str__ = __repr__


class DexFactory(Enum):
    UniswapV2 = auto()  # 0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f

    @property
    def weth(self):
        # TODO: handle by chain_id
        return "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"

    async def get_taxes(
        self,
        token0_address,
        token1_address=None,
        chain_id=1,
    ):
        if not token1_address:
            token1_address = self.weth
        client = _force_get_global_client()
        taxes = await client.prices.get_taxes(
            token0_address,
            token1_address,
            chain_id=chain_id,
        )
        return taxes

    @singledispatchmethod
    async def get_price(
        self,
        token_address: ChecksumAddress,
    ):
        client = _force_get_global_client()
        routes = await client.prices.get_routes(token_address)
        return [
            DexRoute(
                path=row["path"],
                pair_addresses=row["pair_addresses"],
                eth_price=TokenAmount(
                    amount=int(row["eth_price"] * 1e18),
                ),
                usdc_price=TokenAmount(
                    amount=int(row["usdc_price"] * 1e6),
                    decimals=6,
                ),
                factory=self,
            )
            for row in routes
        ]

    @get_price.register(Token)
    async def _(
        self,
        token: Token,
    ):
        return await self.get_price(token.address)

    async def get_pair_info(
        self,
        pair_address: HexAddress,
        force_checksum: bool = True,
        chain_id: int = 1,
    ):
        client = _force_get_global_client()
        pair_info = await client.prices.get_pair_info(
            pair_address,
            force_checksum=force_checksum,
            chain_id=chain_id,
        )
        token0 = Token(**pair_info["token0"])
        token1 = Token(**pair_info["token1"])
        return DexPair(
            factory_address=pair_info["factoryAddress"],
            token0=token0,
            token1=token1,
            address=pair_info["pairAddress"],
            index=pair_info["index"],
            fee=pair_info["feePercentage"],
            network=Network(pair_info["chainId"]),
            block_number=pair_info["blockNumber"],
            transaction_hash=pair_info["transactionHash"],
            factory=self,
        )

    async def get_pairs(
        self,
        token: Token,
    ) -> list["DexPair"]:
        client = _force_get_global_client()
        pairs = await client.prices.get_token_pairs(
            token_address=token.address,
            chain_id=token.network.value,
        )
        return [
            DexPair(
                factory_address=row["factoryAddress"],
                token0=Token(**row["token0"]),
                token1=Token(**row["token1"]),
                address=row["pairAddress"],
                index=row["index"],
                fee=row["feePercentage"],
                network=Network(row["chainId"]),
                block_number=row["blockNumber"],
                transaction_hash=row["transactionHash"],
                factory=self,
            )
            for row in pairs
        ]

    async def swap(
        self,
        wallet: Wallet,
        network: Network,
        path: list[ChecksumAddress],
        amount_in: int,
        slippage_percent: float,
        priority_fee: int,
        is_private: bool,
    ):
        raise NotImplementedError()


class SwapInterval(BaseModel):
    start_time: datetime
    open: float
    close: float
    tx_count: int
    min: float
    max: float
    prev_close: float


class SwapHistory(BaseModel):
    pair: "DexPair"
    intervals: list[SwapInterval]

    @property
    def timestamps(self):
        return [s.start_time for s in self.intervals]

    @property
    def opens(self):
        return [s.prev_close for s in self.intervals]

    @property
    def closes(self):
        return [s.close for s in self.intervals]

    @property
    def mins(self):
        return [s.min for s in self.intervals]

    @property
    def maxs(self):
        return [s.max for s in self.intervals]

    def __repr__(self):
        return f"<SwapHistory: {self.pair.address}>"

    __str__ = __repr__


class DexPair(BaseModel):
    factory_address: ChecksumAddress
    token0: Token
    token1: Token
    address: ChecksumAddress
    index: int

    fee: Optional[float]

    network: Network
    block_number: int

    transaction_hash: HexStr
    factory: DexFactory

    async def get_liquidity(self, block_number: Optional[int] = None):
        client = _force_get_global_client()
        response = await client.prices.get_liquidity(
            token_address=self.address,
            chain_id=self.network.value,
            block_number=block_number,
        )
        return Liquidity(
            token0_balance=response["balances"]["token0"]["amount"],
            token0_price=response["balances"]["token0"]["price"],
            token1_balance=response["balances"]["token1"]["amount"],
            token1_price=response["balances"]["token1"]["price"],
            pair=self,
        )

    async def get_taxes(self):
        return await self.factory.get_taxes(
            self.token0.address,
            self.token1.address,
        )

    async def swap_history(
        self, use_token0: bool = True, start_time=None, end_time=None
    ):
        client = _force_get_global_client()
        feed = await client.prices.load_feed(
            self.address,
            use_token0=use_token0,
        )
        response = []
        for row in sorted(feed.split("\n"))[1:]:
            (
                interval,
                open,
                close,
                min,
                max,
                min_block,
                max_block,
                num_tx,
                prev_close,
            ) = row.split(",")
            response.append(
                SwapInterval(
                    start_time=datetime.strptime(interval, "%Y-%m-%d %H:%M:%S+00:00"),
                    open=float(open),
                    close=float(close),
                    min=float(min),
                    max=float(max),
                    tx_count=int(num_tx),
                    prev_close=float(prev_close if prev_close != "None" else open),
                )
            )
        return SwapHistory(pair=self, intervals=response)

    async def swap(
        self,
        wallet: Wallet,
        amount: int,
        priority_fee: int,
        is_private: bool,
        use_token0: bool,
    ):
        raise NotImplementedError()

    async def honeypot(self) -> bool:
        raise NotImplementedError()

    def __repr__(self):
        return f"<DexPair: {self.token0.symbol}, {self.token1.symbol}>"

    __str__ = __repr__



from functools import singledispatchmethod
from typing import Optional
from uuid import UUID

from eth_typing import ChecksumAddress
from pydantic import BaseModel, Field

from ..utils.client import _force_get_global_client, _set_global_client
from .wallet import Wallet
from .user import User


class Application(BaseModel):
    """
    This represents an Application, which allows a builder to use the SDK.
    """

    id: UUID = Field()
    name: str
    type: str
    tier: str
    api_key: str = Field(alias="apiKey")
    swap_fee: int = Field(alias="swapFee")
    fee_collection_amount: int = Field(alias="feeCollectionAmount")
    request_count: int = Field(alias="requestCount")

    # types
    owner: Optional[User] = Field()
    app_wallet: Optional[Wallet] = Field(alias="appWallet")

    @classmethod
    async def load(self, api_key: Optional[str] = None):
        """
        Loads an instance of the current empyrealSDK user's application.
        If an `api_key` is provided, this will create a new global client.
        Otherwise, the currently set application from the global context
        is loaded.

        Token :class:`empyrealSDK.types.Token`
        """
        from empyrealSDK import EmpyrealSDK

        if api_key:
            new_client = EmpyrealSDK(api_key)
            _set_global_client(new_client)
            return new_client

        client: EmpyrealSDK = _force_get_global_client()
        return await client.app.info()

    async def update_swap_fee(self, swap_fee: float):
        """
        Update your applications swap fee.
        Swap Fee must be set less than 2%.
        """

        client = _force_get_global_client()
        if swap_fee > 0.02:
            raise ValueError("Swap Fee must be less than 2%")
        response = await client.app.update(
            swap_fee=int(swap_fee * 1_000_000),
        )
        obj = response.json()
        for key in obj["updates"]:
            setattr(self, key, obj["updates"][key])
        return self

    @singledispatchmethod
    async def update_app_wallet(self, wallet: Wallet):
        """
        Update your applications swap fee.
        Swap Fee must be set less than 2%.
        """

        client = _force_get_global_client()
        return await client.app.update(
            app_wallet_id=wallet.id,
        )

    @update_app_wallet.register(str)
    async def _(self, wallet_address: ChecksumAddress):
        """
        Update your fee recipient wallet.
        """

        client = _force_get_global_client()
        wallet = await Wallet.load(wallet_address)

        return await client.app.update(
            app_wallet_id=wallet.id,
        )

    async def refresh_api_key(self):
        """
        Update your applications swap fee.
        Swap Fee must be set less than 2%.
        """

        client = _force_get_global_client()
        new_api_key = await client.app.update_api_key()
        self.api_key = new_api_key
        return self



from empyrealSDK.utils import RequestHelpers


class PingResource(RequestHelpers):
    async def say_hi(self):
        """ping method to test connection"""
        response = await self._get("infrastructure/ping")
        return response.json()



# from typing import Optional
# from uuid import UUID
from httpx import Response

from empyrealSDK.utils import RequestHelpers
from empyrealSDK.exc import handle_response_error


class UserResource(RequestHelpers):
    async def get_from_telegram(self, telegram_id: str) -> Response:
        response = await self._get("users/telegram", params={"id": str(telegram_id)})
        handle_response_error(response)
        return response

    # async def get_grants(
    #     self,
    #     user_id: UUID,
    # ) -> Optional[Grant]:
    #     """Get the wallets and wallet types that have been approved by a user
    #        for the current app.

    #     A grant consists of a Grant object with additional GrantTypes.  The
    #     Grant is for the user, which can be pulled to immediately disable
    #     access to all wallets.  This is in contrast to the GrantType, which
    #     is for more granular control over a specific feature.  Initially,
    #     effort will be concentrated on a broad `GrantAllowType.allow_all`,
    #     which provides full access.

    #     Note:
    #         This naming convention may change, I'm on the fence...

    #     Returns:
    #         Grant: if grant is found for user
    #         None: If no grant has been made yet
    #     """
    #     response = await self._get(
    #         "grant/list",
    #         params={"userId": str(user_id)},
    #     )
    #     if response.json():
    #         return Grant(**response.json())
    #     else:
    #         return None

    # async def request_grant_from_user(
    #     self,
    #     wallet_id: UUID,
    #     grant_types: Optional[list[GrantType]] = None,
    # ):
    #     """Request a user to grant you access to their wallet.
    #     This will create a pending request in the user's telegram to accept
    #     your request for access.

    #     The number of grant request an app can send will be controlled based
    #     on their success rate to prevent spamming.

    #     Returns:
    #         bool: success status of request
    #     """
    #     response = await self._put(
    #         "grant/",
    #         json={
    #             "walletId": str(wallet_id),
    #             "grantType": (
    #                 [dict(gt) for gt in grant_types] if grant_types else None,
    #             ),
    #         },
    #     )
    #     return response.status_code == 200



from enum import Enum


class Network(Enum):
    """
    An enum representing the different Networks supported by the SDK
    """

    Ethereum = 1
    # Arbitrum = 42161

    @property
    def chain_id(self):
        """Returns the Chain ID for a network"""
        return self.value



from typing import Mapping
from uuid import UUID

from empyrealSDK.utils import RequestHelpers
from empyrealSDK.types.vault import VaultType, Vault


class VaultResource(RequestHelpers):
    async def get_all(
        self,
    ):
        """Get all vaults for an app

        This will show the general data for each vault

        :return: a list of `Vault`s
        """
        response = await self._get("vault/")
        return [Vault(**v) for v in response.json()["vaults"]]

    async def get_user_positions(
        self,
        user_id: UUID,
    ):
        """Get all vault positions for a user in your app

        This can be used to inspect a user and show them their current
        balance, or to help a user make determinations about how to
        allocate their escrowed funds across different positions.

        :return: A list of `VaultPosition`'s
        """

        response = await self._get(
            "vault/positions",
            params={"userId": str(user_id)},
        )

        return response.json()

    async def make_new_app_vault(
        self,
        token_id: UUID,
        type: VaultType,
        vault_name: str,
        description: str,
    ) -> Mapping[str, str]:
        """Creates a new app vault.  Currently there are two vault types:
            Bank: This is equivalent to wrapping tokens.  Dividends can be
                  issued to all users based on ownership share, but the user
                  will maintain their deposit regardless of what happens to
                  other accounts.
            Vault: This allows for shared earnings and losses distributed by
                   ownership stake in the vault.  This is designed for
                   applications that want to have a shared stake.

        ```python
        await emp_sdk.vault.make_new_app_vault(
            token.id,
            VaultType.bank,
            "UserFunds",
            "allows users to wrap funds into app",
        )
        await emp_sdk.vault.make_new_app_vault(
            token.id,
            VaultType.vault,
            "VolatileFunds",
            "allows users to buy shares of a volatile asset",
        )
        ```

        Returns:
            UUID: the new vault's UUID
        """
        response = await self._post(
            "vault/",
            json={
                "tokenId": str(token_id),
                "type": type.value,
                "vaultName": vault_name,
                "description": description,
            },
        )

        return response.json()



from enum import IntEnum, auto
from typing import Any, Optional, Union
from uuid import UUID

from eth_typing import HexStr
from pydantic import BaseModel, Field

from ..utils.client import _force_get_global_client
from .wallet import Wallet


class UserType(IntEnum):
    telegram = auto()
    unclassified = auto()


class User(BaseModel):
    """
    A User of an application.  This is mainly used to organize wallets
    and actions to a single owner.  Currently the telegram id identity is
    used to manage users, but they can be created without an originating
    telegram account, and it is mainly used as a convenience for tracking.
    """

    id: UUID
    type: Optional[UserType] = None
    name: str
    telegram_id: Optional[str] = Field(alias="telegramId")
    is_new_user: bool = Field(alias="isNewUser")
    metadata: dict[Any, Any] = Field(default={})

    @classmethod
    async def load(cls, telegram_id: Union[str, int]):
        client = _force_get_global_client()
        response = await client.user.get_from_telegram(telegram_id)
        return cls(**response.json())

    @classmethod
    async def create(cls, name: str):
        raise NotImplementedError()

    async def get_app_wallets(self):
        client = _force_get_global_client()
        response = await client.wallet.get_app_wallets(self.id)
        return [Wallet(**row) for row in response.json()]

    async def make_wallet(self, name: str, private_key: Optional[HexStr] = None):
        client = _force_get_global_client()
        response = await client.wallet.make_wallet(
            self.id,
            name,
            private_key=private_key,
        )
        return Wallet(**response.json())

    def __repr__(self):
        return f"<User: {self.name}>"

    __str__ = __repr__



from typing import Literal, Optional, Union
from uuid import UUID

from eth_typing import ChecksumAddress, HexAddress, HexStr
from httpx import Response

from empyrealSDK.utils import RequestHelpers
from empyrealSDK.exc import handle_response_error


class TokenResource(RequestHelpers):
    async def lookup(self, token_address: HexAddress, chain_id: int = 1) -> Response:
        """Looks up a token's info in the SDK given it's address and chain id.
        This is useful for finding a token's id in the application.
        Using an ID instead of address and chain_id for the SDK simplifies
        the vault logic.

        Args:
            token_address: HexAddress a string in Hex String format
            chain_id: an int for the Chain ID

        Returns:
            Token
        """
        response = await self._get(
            "token/lookup",
            params={
                "address": token_address,
                "chainId": chain_id,
            },
        )
        handle_response_error(response)
        return response

    async def transfer(
        self,
        token_id: UUID,
        from_wallet_id: UUID,
        recipient_address: HexAddress,
        amount: int,
        gas_price: Optional[int] = None,
    ) -> HexStr:
        response = await self._get(
            "token/transfer",
            params={
                "walletId": str(from_wallet_id),
                "tokenId": str(token_id),
                "recipientAddress": recipient_address,
                "amount": amount,
                "gasPrice": gas_price,
            },
        )
        return response.json()

    async def balance_of(
        self,
        token_address: HexAddress,
        wallet_address: HexAddress,
        chain_id: int = 1,
        block_num: Union[int, Literal["latest"]] = "latest",
    ) -> int:
        response = await self._put(
            "token/balance",
            json={
                "tokenAddress": token_address,
                "ownerAddress": wallet_address,
                "chainId": chain_id,
                "block": block_num,
            },
        )
        handle_response_error(response)
        return response.json()["balance"]

    async def allowance(
        self,
        token_address: ChecksumAddress,
        owner_address: ChecksumAddress,
        spender_address: ChecksumAddress,
        chain_id: int = 1,
        block_num: Optional[Union[int, Literal["latest"]]] = "latest",
    ) -> int:
        response = await self._get(
            "token/allowance",
            params={
                "tokenAddress": token_address,
                "owner": owner_address,
                "spender": spender_address,
                "chainId": chain_id,
                "block": block_num,
            },
        )
        handle_response_error(response)
        return response.json()["allowance"]

    async def approve(
        self,
        token_id: UUID,
        wallet_id: UUID,
        spender_address: ChecksumAddress,
        chain_id: int = 1,
        amount: int = int(2**256 - 1),
        priority_fee: int = 0,
    ):
        response = await self._post(
            "token/approve",
            json={
                "tokenId": token_id,
                "walletId": wallet_id,
                "spender": spender_address,
                "chainId": chain_id,
                "amount": amount,
                "priorityFee": priority_fee,
            },
        )
        handle_response_error(response)
        return response.json()["txHash"]



import gzip
from typing import Optional

from eth_typing import HexAddress, ChecksumAddress
from eth_utils.address import to_checksum_address

from empyrealSDK.exc import handle_response_error
from empyrealSDK.utils import RequestHelpers


class PriceResource(RequestHelpers):
    async def get_taxes(
        self,
        token_address: ChecksumAddress,
        pair_token_address: ChecksumAddress,
        dex: str = "uniswap",
        chain_id: int = 1,
    ):
        response = await self._put(
            "token/taxes",
            json={
                "tokenAddress": token_address,
                "pairToken": pair_token_address,
                "chainId": chain_id,
                "dex": dex,
            },
        )
        handle_response_error(response)
        return response.json()

    async def get_routes(
        self,
        token_address: ChecksumAddress,
    ):
        response = await self._get(
            "dex/routes",
            params={
                "startToken": token_address,
            },
        )
        handle_response_error(response)
        return response.json()

    async def get_pair_info(
        self,
        pair_address: HexAddress,
        force_checksum: bool = True,
        chain_id: int = 1,
    ):
        """
        Get information about a specific pair addresss
        """

        if force_checksum:
            pair_address = to_checksum_address(pair_address)
        response = await self._get(
            "dex/pair",
            params={
                "pairAddress": pair_address,
                "chainId": chain_id,
            },
        )
        return response.json()

    async def get_token_pairs(
        self,
        token_address: HexAddress,
        chain_id: int = 1,
        force_checksum: bool = True,
    ):
        if force_checksum:
            token_address = to_checksum_address(token_address)
        response = await self._get(
            "dex/pairs",
            params={
                "tokenAddress": token_address,
                "chainId": chain_id,
            },
        )
        return response.json()["pairs"]

    async def get_liquidity(
        self,
        token_address: HexAddress,
        chain_id: int = 1,
        force_checksum: bool = True,
        block_number: Optional[int] = None,
    ):
        if force_checksum:
            token_address = to_checksum_address(token_address)
        response = await self._put(
            "dex/liquidity",
            json={
                "pairAddress": token_address,
                "chainId": chain_id,
                "blockNumber": block_number,
            },
        )
        handle_response_error(response)
        return response.json()

    async def load_feed(
        self,
        pair_address: ChecksumAddress,
        use_token0: bool = True,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
    ):
        response = await self._get(
            "price/",
            params={
                "pairAddress": pair_address,
                "useToken0": use_token0,
            },
        )
        handle_response_error(response)
        return gzip.decompress(response.content).decode("utf-8")
